@@ -960,7 +960,7 @@ public class BcCertTest
         {
             ASN1Enumerated reasonCode = ASN1Enumerated.getInstance(ext.getParsedValue());
 
-            if (reasonCode.getValue().intValue() != CRLReason.privilegeWithdrawn)
+            if (reasonCode.intValueExact() != CRLReason.privilegeWithdrawn)
             {
                 fail("CRL entry reasonCode wrong");
             }
@@ -1041,7 +1041,7 @@ public class BcCertTest
         {
             ASN1Enumerated   reasonCode = ASN1Enumerated.getInstance(ext.getParsedValue());
 
-            if (reasonCode.getValue().intValue() != CRLReason.privilegeWithdrawn)
+            if (reasonCode.intValueExact() != CRLReason.privilegeWithdrawn)
             {
                 fail("CRL entry reasonCode wrong");
             }
@@ -1123,7 +1123,7 @@ public class BcCertTest
         {
             ASN1Enumerated   reasonCode = ASN1Enumerated.getInstance(ext.getParsedValue());
 
-            if (reasonCode.getValue().intValue() != CRLReason.privilegeWithdrawn)
+            if (reasonCode.intValueExact() != CRLReason.privilegeWithdrawn)
             {
                 fail("CRL entry reasonCode wrong");
             }
@@ -1167,7 +1167,7 @@ public class BcCertTest
                 {
                     ASN1Enumerated reasonCode = ASN1Enumerated.getInstance(extn.getParsedValue());
 
-                    if (reasonCode.getValue().intValue() != CRLReason.privilegeWithdrawn)
+                    if (reasonCode.intValueExact() != CRLReason.privilegeWithdrawn)
                     {
                         fail("CRL entry reasonCode wrong on recheck");
                     }
