@@ -5,6 +5,9 @@ import java.nio.ByteBuffer;
 import java.security.Principal;
 import java.security.cert.CertificateException;
 import java.security.cert.X509Certificate;
+import java.util.List;
+import java.util.logging.Level;
+import java.util.logging.Logger;
 
 import javax.net.ssl.SSLEngine;
 import javax.net.ssl.SSLEngineResult;
@@ -13,7 +16,9 @@ import javax.net.ssl.SSLEngineResult.Status;
 import javax.net.ssl.SSLException;
 import javax.net.ssl.SSLParameters;
 import javax.net.ssl.SSLSession;
+import javax.net.ssl.X509ExtendedKeyManager;
 
+import org.bouncycastle.jsse.BCApplicationProtocolSelector;
 import org.bouncycastle.jsse.BCExtendedSSLSession;
 import org.bouncycastle.jsse.BCSSLConnection;
 import org.bouncycastle.jsse.BCSSLEngine;
@@ -36,13 +41,16 @@ class ProvSSLEngine
     extends SSLEngine
     implements BCSSLEngine, ProvTlsManager
 {
-    protected final ProvSSLContextSpi context;
+    private static final Logger LOG = Logger.getLogger(ProvSSLEngine.class.getName());
+
     protected final ContextData contextData;
     protected final ProvSSLParameters sslParameters;
 
     protected boolean enableSessionCreation = true;
-    protected boolean useClientMode = false;
+    protected boolean useClientMode = true;
+    protected boolean useClientModeSet = false;
 
+    protected boolean closedEarly = false;
     protected boolean initialHandshakeBegun = false;
     protected HandshakeStatus handshakeStatus = HandshakeStatus.NOT_HANDSHAKING; 
     protected TlsProtocol protocol = null;
@@ -52,27 +60,17 @@ class ProvSSLEngine
 
     protected SSLException deferredException = null;
 
-    protected ProvSSLEngine(ProvSSLContextSpi context, ContextData contextData)
+    protected ProvSSLEngine(ContextData contextData)
     {
-        super();
-
-        this.context = context;
-        this.contextData = contextData;
-        this.sslParameters = context.getDefaultParameters(!useClientMode);
+        this(contextData, null, -1);
     }
 
-    protected ProvSSLEngine(ProvSSLContextSpi context, ContextData contextData, String host, int port)
+    protected ProvSSLEngine(ContextData contextData, String host, int port)
     {
         super(host, port);
 
-        this.context = context;
         this.contextData = contextData;
-        this.sslParameters = context.getDefaultParameters(!useClientMode);
-    }
-
-    public ProvSSLContextSpi getContext()
-    {
-        return context;
+        this.sslParameters = contextData.getContext().getDefaultSSLParameters(useClientMode);
     }
 
     public ContextData getContextData()
@@ -84,6 +82,14 @@ class ProvSSLEngine
     public synchronized void beginHandshake()
         throws SSLException
     {
+        if (!useClientModeSet)
+        {
+            throw new IllegalStateException("Client/Server mode must be set before the handshake can begin");
+        }
+        if (closedEarly)
+        {
+            throw new SSLException("Connection is already closed");
+        }
         if (initialHandshakeBegun)
         {
             throw new UnsupportedOperationException("Renegotiation not supported");
@@ -98,7 +104,7 @@ class ProvSSLEngine
                 TlsClientProtocol clientProtocol = new TlsClientProtocol();
                 this.protocol = clientProtocol;
 
-                ProvTlsClient client = new ProvTlsClient(this, sslParameters.copy());
+                ProvTlsClient client = new ProvTlsClient(this, sslParameters);
                 this.protocolPeer = client;
 
                 clientProtocol.connect(client);
@@ -109,7 +115,7 @@ class ProvSSLEngine
                 TlsServerProtocol serverProtocol = new TlsServerProtocol();
                 this.protocol = serverProtocol;
 
-                ProvTlsServer server = new ProvTlsServer(this, sslParameters.copy());
+                ProvTlsServer server = new ProvTlsServer(this, sslParameters);
                 this.protocolPeer = server;
 
                 serverProtocol.accept(server);
@@ -130,7 +136,7 @@ class ProvSSLEngine
     {
         try
         {
-            contextData.getX509TrustManager().checkClientTrusted(chain, authType, this);
+            contextData.getX509TrustManager().checkClientTrusted(chain.clone(), authType, this);
         }
         catch (CertificateException e)
         {
@@ -142,7 +148,7 @@ class ProvSSLEngine
     {
         try
         {
-            contextData.getX509TrustManager().checkServerTrusted(chain, authType, this);
+            contextData.getX509TrustManager().checkServerTrusted(chain.clone(), authType, this);
         }
         catch (CertificateException e)
         {
@@ -150,51 +156,78 @@ class ProvSSLEngine
         }
     }
 
-    public String chooseClientAlias(String[] keyType, Principal[] issuers)
+    public ProvX509Key chooseClientKey(String[] keyTypes, Principal[] issuers)
     {
-        return contextData.getX509KeyManager().chooseEngineClientAlias(keyType, issuers, this);
+        X509ExtendedKeyManager x509KeyManager = getContextData().getX509KeyManager();
+        String alias = x509KeyManager.chooseEngineClientAlias(keyTypes, issuers, this);
+        return ProvX509Key.from(x509KeyManager, alias);
     }
 
-    public String chooseServerAlias(String keyType, Principal[] issuers)
+    public ProvX509Key chooseServerKey(String keyType, Principal[] issuers)
     {
-        return contextData.getX509KeyManager().chooseEngineServerAlias(keyType, issuers, this);
+        X509ExtendedKeyManager x509KeyManager = getContextData().getX509KeyManager();
+        String alias = x509KeyManager.chooseEngineServerAlias(keyType, issuers, this);
+        return ProvX509Key.from(x509KeyManager, alias);
     }
 
     @Override
     public synchronized void closeInbound()
         throws SSLException
     {
-        // TODO How to behave when protocol is still null?
-        try
+        if (closedEarly)
         {
-            protocol.closeInput();
+            // SSLEngine already closed before any handshake attempted
         }
-        catch (IOException e)
+        else if (null == protocol)
         {
-            throw new SSLException(e);
+            this.closedEarly = true;
+        }
+        else
+        {
+            try
+            {
+                protocol.closeInput();
+            }
+            catch (IOException e)
+            {
+                throw new SSLException(e);
+            }
         }
     }
 
     @Override
     public synchronized void closeOutbound()
     {
-        // TODO How to behave when protocol is still null?
-        try
+        if (closedEarly)
         {
-            protocol.close();
+            // SSLEngine already closed before any handshake attempted
         }
-        catch (IOException e)
+        else if (null == protocol)
         {
-           // TODO[logging] 
+            this.closedEarly = true;
+        }
+        else
+        {
+            try
+            {
+                protocol.close();
+            }
+            catch (IOException e)
+            {
+                LOG.log(Level.WARNING, "Failed to close outbound", e);
+            }
         }
     }
 
-    // @Override from JDK 9
-    public String getApplicationProtocol()
+    // An SSLEngine method from JDK 9, but also a BCSSLEngine method
+    public synchronized String getApplicationProtocol()
     {
-        BCSSLConnection connection = getConnection();
+        return null == connection ? null : connection.getApplicationProtocol();
+    }
 
-        return connection == null ? null : connection.getApplicationProtocol();
+    public synchronized BCApplicationProtocolSelector<SSLEngine> getBCHandshakeApplicationProtocolSelector()
+    {
+        return sslParameters.getEngineAPSelector();
     }
 
     public synchronized BCExtendedSSLSession getBCHandshakeSession()
@@ -202,6 +235,11 @@ class ProvSSLEngine
         return handshakeSession;
     }
 
+    public BCExtendedSSLSession getBCSession()
+    {
+        return getSessionImpl();
+    }
+
     public synchronized BCSSLConnection getConnection()
     {
         return connection;
@@ -231,7 +269,13 @@ class ProvSSLEngine
         return enableSessionCreation;
     }
 
-    @Override
+    // An SSLEngine method from JDK 9, but also a BCSSLEngine method
+    public synchronized String getHandshakeApplicationProtocol()
+    {
+        return null == handshakeSession ? null : handshakeSession.getApplicationProtocol();
+    }
+
+    // An SSLEngine method from JDK 7
     public synchronized SSLSession getHandshakeSession()
     {
         return null == handshakeSession ? null : handshakeSession.getExportSSLSession();
@@ -255,14 +299,12 @@ class ProvSSLEngine
     }
 
     @Override
-    public synchronized SSLSession getSession()
+    public SSLSession getSession()
     {
-        ProvSSLSession sslSession = (null == connection) ? ProvSSLSession.NULL_SESSION : connection.getSession();
-
-        return sslSession.getExportSSLSession();
+        return getSessionImpl().getExportSSLSession();
     }
 
-    @Override
+    // An SSLEngine method from JDK 6
     public synchronized SSLParameters getSSLParameters()
     {
         return SSLParametersUtil.getSSLParameters(sslParameters);
@@ -271,13 +313,13 @@ class ProvSSLEngine
     @Override
     public synchronized String[] getSupportedCipherSuites()
     {
-        return context.getSupportedCipherSuites();
+        return contextData.getContext().getSupportedCipherSuites();
     }
 
     @Override
     public synchronized String[] getSupportedProtocols()
     {
-        return context.getSupportedProtocols();
+        return contextData.getContext().getSupportedProtocols();
     }
 
     @Override
@@ -295,13 +337,36 @@ class ProvSSLEngine
     @Override
     public synchronized boolean isInboundDone()
     {
-        return protocol != null && protocol.isClosed();
+        return closedEarly || (null != protocol && protocol.isClosed());
     }
 
     @Override
     public synchronized boolean isOutboundDone()
     {
-        return protocol != null && protocol.isClosed() && protocol.getAvailableOutputBytes() < 1;
+        return closedEarly || (null != protocol && protocol.isClosed() && protocol.getAvailableOutputBytes() < 1);
+    }
+
+    public synchronized void setBCHandshakeApplicationProtocolSelector(BCApplicationProtocolSelector<SSLEngine> selector)
+    {
+        sslParameters.setEngineAPSelector(selector);
+    }
+
+    public synchronized void setBCSessionToResume(BCExtendedSSLSession session)
+    {
+        if (null == session)
+        {
+            throw new NullPointerException("'session' cannot be null");
+        }
+        if (!(session instanceof ProvSSLSession))
+        {
+            throw new IllegalArgumentException("Session-to-resume must be a session returned from 'getBCSession'");
+        }
+        if (initialHandshakeBegun)
+        {
+            throw new IllegalArgumentException("Session-to-resume cannot be set after the handshake has begun");
+        }
+
+        sslParameters.setSessionToResume((ProvSSLSession)session);
     }
 
     @Override
@@ -333,7 +398,7 @@ class ProvSSLEngine
         SSLParametersUtil.setParameters(this.sslParameters, parameters);
     }
 
-    @Override
+    // An SSLEngine method from JDK 6
     public synchronized void setSSLParameters(SSLParameters sslParameters)
     {
         SSLParametersUtil.setSSLParameters(this.sslParameters, sslParameters);
@@ -344,15 +409,17 @@ class ProvSSLEngine
     {
         if (initialHandshakeBegun)
         {
-            throw new IllegalArgumentException("Mode cannot be changed after the initial handshake has begun");
+            throw new IllegalArgumentException("Client/Server mode cannot be changed after the handshake has begun");
         }
 
         if (this.useClientMode != useClientMode)
         {
-            context.updateDefaultProtocols(sslParameters, !useClientMode);
+            contextData.getContext().updateDefaultSSLParameters(sslParameters, useClientMode);
 
             this.useClientMode = useClientMode;
         }
+
+        this.useClientModeSet = true;
     }
 
     @Override
@@ -640,9 +707,14 @@ class ProvSSLEngine
 
     public synchronized void notifyHandshakeComplete(ProvSSLConnection connection)
     {
-        if (null != handshakeSession && !handshakeSession.isValid())
+        if (null != handshakeSession)
         {
-            connection.getSession().invalidate();
+            if (!handshakeSession.isValid())
+            {
+                connection.getSession().invalidate();
+            }
+
+            handshakeSession.getJsseSecurityParameters().clear();
         }
 
         this.handshakeSession = null;
@@ -654,6 +726,16 @@ class ProvSSLEngine
         this.handshakeSession = handshakeSession;
     }
 
+    public synchronized String selectApplicationProtocol(List<String> protocols)
+    {
+        return sslParameters.getEngineAPSelector().select(this, protocols);
+    }
+
+    ProvSSLSession getSessionImpl()
+    {
+        return null == connection ? ProvSSLSession.NULL_SESSION : connection.getSession();
+    }
+
     private RecordPreview getRecordPreview(ByteBuffer src)
         throws IOException
     {
