@@ -6,6 +6,7 @@ import java.io.IOException;
 import java.util.Hashtable;
 import java.util.Vector;
 
+import org.bouncycastle.tls.crypto.TlsSecret;
 import org.bouncycastle.util.Arrays;
 
 public class DTLSServerProtocol
@@ -30,6 +31,12 @@ public class DTLSServerProtocol
 
     public DTLSTransport accept(TlsServer server, DatagramTransport transport)
         throws IOException
+    {
+        return accept(server, transport, null);
+    }
+
+    public DTLSTransport accept(TlsServer server, DatagramTransport transport, DTLSRequest request)
+        throws IOException
     {
         if (server == null)
         {
@@ -49,13 +56,12 @@ public class DTLSServerProtocol
         SecurityParameters securityParameters = state.serverContext.getSecurityParametersHandshake();
         securityParameters.extendedPadding = server.shouldUseExtendedPadding();
 
-        DTLSRecordLayer recordLayer = new DTLSRecordLayer(transport, server, ContentType.handshake);
-
-        // TODO Need to handle sending of HelloVerifyRequest without entering a full connection
+        DTLSRecordLayer recordLayer = new DTLSRecordLayer(state.serverContext, state.server, transport);
+        server.notifyCloseHandle(recordLayer);
 
         try
         {
-            return serverHandshake(state, recordLayer);
+            return serverHandshake(state, recordLayer, request);
         }
         catch (TlsFatalAlert fatalAlert)
         {
@@ -84,24 +90,34 @@ public class DTLSServerProtocol
         invalidateSession(state);
     }
 
-    protected DTLSTransport serverHandshake(ServerHandshakeState state, DTLSRecordLayer recordLayer)
-        throws IOException
+    protected DTLSTransport serverHandshake(ServerHandshakeState state, DTLSRecordLayer recordLayer,
+        DTLSRequest request) throws IOException
     {
         SecurityParameters securityParameters = state.serverContext.getSecurityParametersHandshake();
-        DTLSReliableHandshake handshake = new DTLSReliableHandshake(state.serverContext, recordLayer);
+        DTLSReliableHandshake handshake = new DTLSReliableHandshake(state.serverContext, recordLayer,
+            state.server.getHandshakeTimeoutMillis(), request);
 
-        DTLSReliableHandshake.Message clientMessage = handshake.receiveMessage();
+        DTLSReliableHandshake.Message clientMessage = null;
 
-        // NOTE: DTLSRecordLayer requires any DTLS version, we don't otherwise constrain this
-//        ProtocolVersion recordLayerVersion = recordLayer.getReadVersion();
-
-        if (clientMessage.getType() == HandshakeType.client_hello)
+        if (null == request)
         {
-            processClientHello(state, clientMessage.getBody());
+            clientMessage = handshake.receiveMessage();
+
+            // NOTE: DTLSRecordLayer requires any DTLS version, we don't otherwise constrain this
+//            ProtocolVersion recordLayerVersion = recordLayer.getReadVersion();
+
+            if (clientMessage.getType() == HandshakeType.client_hello)
+            {
+                processClientHello(state, clientMessage.getBody());
+            }
+            else
+            {
+                throw new TlsFatalAlert(AlertDescription.unexpected_message);
+            }
         }
         else
         {
-            throw new TlsFatalAlert(AlertDescription.unexpected_message);
+            processClientHello(state, request.getClientHello());
         }
 
         /*
@@ -117,6 +133,7 @@ public class DTLSServerProtocol
 
             state.tlsSession = TlsUtils.importSession(securityParameters.getSessionID(), null);
             state.sessionParameters = null;
+            state.sessionMasterSecret = null;
         }
 
         {
@@ -132,7 +149,7 @@ public class DTLSServerProtocol
             handshake.sendMessage(HandshakeType.server_hello, serverHelloBody);
         }
 
-        handshake.notifyHelloComplete();
+        handshake.getHandshakeHash().notifyPRFDetermined();
 
         Vector serverSupplementalData = state.server.getServerSupplementalData();
         if (serverSupplementalData != null)
@@ -142,8 +159,7 @@ public class DTLSServerProtocol
         }
 
         state.keyExchange = TlsUtils.initKeyExchangeServer(state.serverContext, state.server);
-
-        state.serverCredentials = TlsProtocol.validateCredentials(state.server.getCredentials());
+        state.serverCredentials = TlsUtils.establishServerCredentials(state.server);
 
         // Server certificate
         {
@@ -190,7 +206,20 @@ public class DTLSServerProtocol
         if (state.serverCredentials != null)
         {
             state.certificateRequest = state.server.getCertificateRequest();
-            if (state.certificateRequest != null)
+
+            if (null == state.certificateRequest)
+            {
+                /*
+                 * For static agreement key exchanges, CertificateRequest is required since
+                 * the client Certificate message is mandatory but can only be sent if the
+                 * server requests it.
+                 */
+                if (!state.keyExchange.requiresCertificateVerify())
+                {
+                    throw new TlsFatalAlert(AlertDescription.internal_error);
+                }
+            }
+            else
             {
                 if (TlsUtils.isTLSv12(state.serverContext) != (state.certificateRequest.getSupportedSignatureAlgorithms() != null))
                 {
@@ -199,11 +228,12 @@ public class DTLSServerProtocol
 
                 state.certificateRequest = TlsUtils.validateCertificateRequest(state.certificateRequest, state.keyExchange);
 
+                TlsUtils.establishServerSigAlgs(securityParameters, state.certificateRequest);
+
+                TlsUtils.trackHashAlgorithms(handshake.getHandshakeHash(), securityParameters.getServerSigAlgs());
+
                 byte[] certificateRequestBody = generateCertificateRequest(state, state.certificateRequest);
                 handshake.sendMessage(HandshakeType.certificate_request, certificateRequestBody);
-
-                TlsUtils.trackHashAlgorithms(handshake.getHandshakeHash(),
-                    state.certificateRequest.getSupportedSignatureAlgorithms());
             }
         }
 
@@ -261,8 +291,7 @@ public class DTLSServerProtocol
             throw new TlsFatalAlert(AlertDescription.unexpected_message);
         }
 
-        TlsHandshakeHash prepareFinishHash = handshake.prepareToFinish();
-        securityParameters.sessionHash = TlsUtils.getCurrentPRFHash(prepareFinishHash);
+        securityParameters.sessionHash = TlsUtils.getCurrentPRFHash(handshake.getHandshakeHash());
 
         TlsProtocol.establishMasterSecret(state.serverContext, state.keyExchange);
         recordLayer.initPendingEpoch(TlsUtils.initCipher(state.serverContext));
@@ -272,10 +301,14 @@ public class DTLSServerProtocol
          * capability (i.e., all certificates except those containing fixed Diffie-Hellman
          * parameters).
          */
-        if (expectCertificateVerifyMessage(state))
         {
-            byte[] certificateVerifyBody = handshake.receiveMessageBody(HandshakeType.certificate_verify);
-            processCertificateVerify(state, certificateVerifyBody, prepareFinishHash);
+            TlsHandshakeHash certificateVerifyHash = handshake.prepareToFinish();
+
+            if (expectCertificateVerifyMessage(state))
+            {
+                byte[] certificateVerifyBody = handshake.receiveMessageBody(HandshakeType.certificate_verify);
+                processCertificateVerify(state, certificateVerifyBody, certificateVerifyHash);
+            }
         }
 
         // NOTE: Calculated exclusive of the actual Finished message from the client
@@ -286,7 +319,7 @@ public class DTLSServerProtocol
         {
             NewSessionTicket newSessionTicket = state.server.getNewSessionTicket();
             byte[] newSessionTicketBody = generateNewSessionTicket(state, newSessionTicket);
-            handshake.sendMessage(HandshakeType.session_ticket, newSessionTicketBody);
+            handshake.sendMessage(HandshakeType.new_session_ticket, newSessionTicketBody);
         }
 
         // NOTE: Calculated exclusive of the Finished message itself
@@ -295,12 +328,14 @@ public class DTLSServerProtocol
 
         handshake.finish();
 
+        state.sessionMasterSecret = securityParameters.getMasterSecret();
+
         state.sessionParameters = new SessionParameters.Builder()
             .setCipherSuite(securityParameters.getCipherSuite())
             .setCompressionAlgorithm(securityParameters.getCompressionAlgorithm())
             .setExtendedMasterSecret(securityParameters.isExtendedMasterSecret())
             .setLocalCertificate(securityParameters.getLocalCertificate())
-            .setMasterSecret(state.serverContext.getCrypto().adoptSecret(securityParameters.getMasterSecret()))
+            .setMasterSecret(state.serverContext.getCrypto().adoptSecret(state.sessionMasterSecret))
             .setNegotiatedVersion(securityParameters.getNegotiatedVersion())
             .setPeerCertificate(securityParameters.getPeerCertificate())
             .setPSKIdentity(securityParameters.getPSKIdentity())
@@ -315,6 +350,8 @@ public class DTLSServerProtocol
 
         state.serverContext.handshakeComplete(state.server, state.tlsSession);
 
+        recordLayer.initHeartbeat(state.heartbeat, HeartbeatMode.peer_allowed_to_send == state.heartbeatPolicy);
+
         return new DTLSTransport(recordLayer);
     }
 
@@ -350,7 +387,7 @@ public class DTLSServerProtocol
 
         ProtocolVersion server_version = state.server.getServerVersion();
         {
-            if (null == server_version || !ProtocolVersion.DTLSv10.isEqualOrEarlierVersionOf(server_version)
+            if (!ProtocolVersion.isSupportedDTLSVersion(server_version)
                 || !ProtocolVersion.contains(context.getClientSupportedVersions(), server_version))
             {
                 throw new TlsFatalAlert(AlertDescription.internal_error);
@@ -363,10 +400,11 @@ public class DTLSServerProtocol
 //
 //            recordLayer.setWriteVersion(legacy_record_version);
             securityParameters.negotiatedVersion = server_version;
+            TlsUtils.negotiatedVersion(context);
         }
 
         securityParameters.serverRandom = TlsProtocol.createRandomBlock(state.server.shouldUseGMTUnixTime(), context);
-        if (!server_version.equals(ProtocolVersion.getLatestDTLS(state.server.getSupportedVersions())))
+        if (!server_version.equals(ProtocolVersion.getLatestDTLS(state.server.getProtocolVersions())))
         {
             TlsUtils.writeDowngradeMarker(server_version, securityParameters.getServerRandom());
         }
@@ -382,6 +420,7 @@ public class DTLSServerProtocol
             }
             securityParameters.cipherSuite = validateSelectedCipherSuite(selectedCipherSuite,
                 AlertDescription.internal_error);
+            TlsUtils.negotiatedCipherSuite(state.serverContext);
         }
 
         state.serverExtensions = TlsExtensionsUtils.ensureExtensionsInitialised(state.server.getServerExtensions());
@@ -421,17 +460,43 @@ public class DTLSServerProtocol
             }
         }
 
-        if (securityParameters.isExtendedMasterSecret())
+        /*
+         * RFC 7627 4. Clients and servers SHOULD NOT accept handshakes that do not use the extended
+         * master secret [..]. (and see 5.2, 5.3)
+         */
         {
-            TlsExtensionsUtils.addExtendedMasterSecretExtension(state.serverExtensions);
+            securityParameters.extendedMasterSecret = state.offeredExtendedMasterSecret
+                && state.server.shouldUseExtendedMasterSecret();
+
+            if (securityParameters.isExtendedMasterSecret())
+            {
+                TlsExtensionsUtils.addExtendedMasterSecretExtension(state.serverExtensions);
+            }
+            else if (state.server.requiresExtendedMasterSecret())
+            {
+                throw new TlsFatalAlert(AlertDescription.handshake_failure);
+            }
+            else if (state.resumedSession && !state.server.allowLegacyResumption())
+            {
+                throw new TlsFatalAlert(AlertDescription.internal_error);
+            }
         }
 
+        // Heartbeats
+        if (null != state.heartbeat || HeartbeatMode.peer_allowed_to_send == state.heartbeatPolicy)
+        {
+            TlsExtensionsUtils.addHeartbeatExtension(state.serverExtensions, new HeartbeatExtension(state.heartbeatPolicy));
+        }
+
+
+
         /*
          * RFC 7301 3.1. When session resumption or session tickets [...] are used, the previous
          * contents of this extension are irrelevant, and only the values in the new handshake
          * messages are considered.
          */
         securityParameters.applicationProtocol = TlsExtensionsUtils.getALPNExtensionServer(state.serverExtensions);
+        securityParameters.applicationProtocolSet = true;
 
         /*
          * TODO RFC 3546 2.3 If [...] the older session is resumed, then the server MUST ignore
@@ -473,29 +538,22 @@ public class DTLSServerProtocol
 
 
 
+        ServerHello serverHello = new ServerHello(legacy_version, securityParameters.getServerRandom(),
+            state.tlsSession.getSessionID(), securityParameters.getCipherSuite(), state.serverExtensions);
+
         ByteArrayOutputStream buf = new ByteArrayOutputStream();
-
-        TlsUtils.writeVersion(legacy_version, buf);
-
-        buf.write(securityParameters.getServerRandom());
-
-        /*
-         * The server may return an empty session_id to indicate that the session will not be cached
-         * and therefore cannot be resumed.
-         */
-        TlsUtils.writeOpaque8(state.tlsSession.getSessionID(), buf);
-
-        TlsUtils.writeUint16(securityParameters.getCipherSuite(), buf);
-
-        TlsUtils.writeUint8(CompressionMethod._null, buf);
-
-        TlsProtocol.writeExtensions(buf, state.serverExtensions);
-
+        serverHello.encode(state.serverContext, buf);
         return buf.toByteArray();
     }
 
     protected void invalidateSession(ServerHandshakeState state)
     {
+        if (state.sessionMasterSecret != null)
+        {
+            state.sessionMasterSecret.destroy();
+            state.sessionMasterSecret = null;
+        }
+
         if (state.sessionParameters != null)
         {
             state.sessionParameters.clear();
@@ -528,7 +586,7 @@ public class DTLSServerProtocol
         notifyClientCertificate(state, clientCertificate);
     }
 
-    protected void processCertificateVerify(ServerHandshakeState state, byte[] body, TlsHandshakeHash prepareFinishHash)
+    protected void processCertificateVerify(ServerHandshakeState state, byte[] body, TlsHandshakeHash handshakeHash)
         throws IOException
     {
         if (state.certificateRequest == null)
@@ -543,66 +601,47 @@ public class DTLSServerProtocol
 
         TlsProtocol.assertEmpty(buf);
 
-        TlsUtils.verifyCertificateVerify(context, state.certificateRequest, clientCertificateVerify, prepareFinishHash);
+        TlsUtils.verifyCertificateVerify(context, state.certificateRequest, clientCertificateVerify, handshakeHash);
     }
 
     protected void processClientHello(ServerHandshakeState state, byte[] body)
         throws IOException
     {
         ByteArrayInputStream buf = new ByteArrayInputStream(body);
+        ClientHello clientHello = ClientHello.parse(buf, NullOutputStream.INSTANCE);
+        processClientHello(state, clientHello);
+    }
 
+    protected void processClientHello(ServerHandshakeState state, ClientHello clientHello)
+        throws IOException
+    {
         // TODO Read RFCs for guidance on the expected record layer version number
-        ProtocolVersion client_version = TlsUtils.readVersion(buf);
-
-        /*
-         * Read the client random
-         */
-        byte[] client_random = TlsUtils.readFully(32, buf);
-
-        byte[] sessionID = TlsUtils.readOpaque8(buf, 0, 32);
-
-        /*
-         * RFC 6347 This specification increases the cookie size limit to 255 bytes for greater
-         * future flexibility. The limit remains 32 for previous versions of DTLS.
-         */
-        int maxCookieLength = ProtocolVersion.DTLSv12.isEqualOrEarlierVersionOf(client_version) ? 255 : 32;
-
-        byte[] cookie = TlsUtils.readOpaque8(buf, 0, maxCookieLength);
-
-        int cipher_suites_length = TlsUtils.readUint16(buf);
-        if (cipher_suites_length < 2 || (cipher_suites_length & 1) != 0)
-        {
-            throw new TlsFatalAlert(AlertDescription.decode_error);
-        }
-
-        /*
-         * NOTE: "If the session_id field is not empty (implying a session resumption request) this
-         * vector must include at least the cipher_suite from that session."
-         */
-        state.offeredCipherSuites = TlsUtils.readUint16Array(cipher_suites_length / 2, buf);
-
-        int compression_methods_length = TlsUtils.readUint8(buf);
-        if (compression_methods_length < 1)
-        {
-            throw new TlsFatalAlert(AlertDescription.illegal_parameter);
-        }
-
-        short[] offeredCompressionMethods = TlsUtils.readUint8Array(compression_methods_length, buf);
+        ProtocolVersion legacy_version = clientHello.getVersion();
+        state.offeredCipherSuites = clientHello.getCipherSuites();
 
         /*
          * TODO RFC 3546 2.3 If [...] the older session is resumed, then the server MUST ignore
          * extensions appearing in the client hello, and send a server hello containing no
          * extensions.
          */
-        state.clientExtensions = TlsProtocol.readExtensions(buf);
+        state.clientExtensions = clientHello.getExtensions();
 
     
     
         TlsServerContextImpl context = state.serverContext;
         SecurityParameters securityParameters = context.getSecurityParametersHandshake();
 
+        if (!legacy_version.isDTLS())
+        {
+            throw new TlsFatalAlert(AlertDescription.illegal_parameter);
+        }
+
+        context.setRSAPreMasterSecretVersion(legacy_version);
+
         context.setClientSupportedVersions(
             TlsExtensionsUtils.getSupportedVersionsExtensionClient(state.clientExtensions));
+
+        ProtocolVersion client_version = legacy_version;
         if (null == context.getClientSupportedVersions())
         {
             if (client_version.isLaterVersionOf(ProtocolVersion.DTLSv12))
@@ -617,40 +656,25 @@ public class DTLSServerProtocol
             client_version = ProtocolVersion.getLatestDTLS(context.getClientSupportedVersions());
         }
 
-        if (null == client_version || !ProtocolVersion.DTLSv10.isEqualOrEarlierVersionOf(client_version))
+        if (!ProtocolVersion.EARLIEST_SUPPORTED_DTLS.isEqualOrEarlierVersionOf(client_version))
         {
-            throw new TlsFatalAlert(AlertDescription.illegal_parameter);
+            throw new TlsFatalAlert(AlertDescription.protocol_version);
         }
 
         context.setClientVersion(client_version);
 
         state.server.notifyClientVersion(context.getClientVersion());
 
-        securityParameters.clientRandom = client_random;
+        securityParameters.clientRandom = clientHello.getRandom();
 
         state.server.notifyFallback(Arrays.contains(state.offeredCipherSuites, CipherSuite.TLS_FALLBACK_SCSV));
 
         state.server.notifyOfferedCipherSuites(state.offeredCipherSuites);
 
-        if (!Arrays.contains(offeredCompressionMethods, CompressionMethod._null))
-        {
-            throw new TlsFatalAlert(AlertDescription.handshake_failure);
-        }
-
         /*
          * TODO[resumption] Check RFC 7627 5.4. for required behaviour 
          */
 
-        /*
-         * RFC 7627 4. Clients and servers SHOULD NOT accept handshakes that do not use the extended
-         * master secret [..]. (and see 5.2, 5.3)
-         */
-        securityParameters.extendedMasterSecret = TlsExtensionsUtils.hasExtendedMasterSecretExtension(state.clientExtensions);
-        if (!securityParameters.isExtendedMasterSecret() && state.server.requiresExtendedMasterSecret())
-        {
-            throw new TlsFatalAlert(AlertDescription.handshake_failure);
-        }
-
         /*
          * RFC 5746 3.6. Server Behavior: Initial Handshake
          */
@@ -694,6 +718,8 @@ public class DTLSServerProtocol
 
         state.server.notifySecureRenegotiation(securityParameters.isSecureRenegotiation());
 
+        state.offeredExtendedMasterSecret = TlsExtensionsUtils.hasExtendedMasterSecretExtension(state.clientExtensions);
+
         if (state.clientExtensions != null)
         {
             // NOTE: Validates the padding extension data, if present
@@ -707,12 +733,25 @@ public class DTLSServerProtocol
              */
             if (TlsUtils.isSignatureAlgorithmsExtensionAllowed(client_version))
             {
-                securityParameters.clientSigAlgs = TlsExtensionsUtils.getSignatureAlgorithmsExtension(state.clientExtensions);
-                securityParameters.clientSigAlgsCert = TlsExtensionsUtils.getSignatureAlgorithmsCertExtension(state.clientExtensions);
+                TlsUtils.establishClientSigAlgs(securityParameters, state.clientExtensions);
             }
 
             securityParameters.clientSupportedGroups = TlsExtensionsUtils.getSupportedGroupsExtension(state.clientExtensions);
 
+            // Heartbeats
+            {
+                HeartbeatExtension heartbeatExtension = TlsExtensionsUtils.getHeartbeatExtension(state.clientExtensions);
+                if (null != heartbeatExtension)
+                {
+                    if (HeartbeatMode.peer_allowed_to_send == heartbeatExtension.getMode())
+                    {
+                        state.heartbeat = state.server.getHeartbeat();
+                    }
+
+                    state.heartbeatPolicy = state.server.getHeartbeatPolicy();
+                }
+            }
+
             state.server.processClientExtensions(state.clientExtensions);
         }
     }
@@ -748,15 +787,19 @@ public class DTLSServerProtocol
         TlsServerContextImpl serverContext = null;
         TlsSession tlsSession = null;
         SessionParameters sessionParameters = null;
+        TlsSecret sessionMasterSecret = null;
         SessionParameters.Builder sessionParametersBuilder = null;
         int[] offeredCipherSuites = null;
         Hashtable clientExtensions = null;
         Hashtable serverExtensions = null;
+        boolean offeredExtendedMasterSecret = false;
         boolean resumedSession = false;
         boolean allowCertificateStatus = false;
         boolean expectSessionTicket = false;
         TlsKeyExchange keyExchange = null;
         TlsCredentials serverCredentials = null;
         CertificateRequest certificateRequest = null;
+        TlsHeartbeat heartbeat = null;
+        short heartbeatPolicy = HeartbeatMode.peer_not_allowed_to_send;
     }
 }
