@@ -44,7 +44,13 @@ public class SRPTlsServer
 
     protected int[] getSupportedCipherSuites()
     {
-        return TlsUtils.getSupportedCipherSuites(context.getCrypto(), DEFAULT_CIPHER_SUITES);
+        return TlsUtils.getSupportedCipherSuites(getCrypto(), DEFAULT_CIPHER_SUITES);
+    }
+
+    public ProtocolVersion getMaximumVersion()
+    {
+        // TODO[tls13] Consider whether this class should offer TLSv13
+        return ProtocolVersion.TLSv12;
     }
 
     public void processClientExtensions(Hashtable clientExtensions) throws IOException
@@ -76,7 +82,7 @@ public class SRPTlsServer
 
     public TlsCredentials getCredentials() throws IOException
     {
-        int keyExchangeAlgorithm = TlsUtils.getKeyExchangeAlgorithm(selectedCipherSuite);
+        int keyExchangeAlgorithm = context.getSecurityParametersHandshake().getKeyExchangeAlgorithm();
 
         switch (keyExchangeAlgorithm)
         {
