@@ -3,23 +3,36 @@ package org.bouncycastle.tls;
 import java.io.ByteArrayInputStream;
 import java.io.ByteArrayOutputStream;
 import java.io.IOException;
+import java.util.Enumeration;
 import java.util.Hashtable;
 import java.util.Vector;
 
+import org.bouncycastle.asn1.ASN1Encoding;
+import org.bouncycastle.asn1.ASN1ObjectIdentifier;
+import org.bouncycastle.asn1.ASN1Primitive;
+import org.bouncycastle.asn1.x500.X500Name;
 import org.bouncycastle.util.Arrays;
 import org.bouncycastle.util.Integers;
 
 public class TlsExtensionsUtils
 {
     public static final Integer EXT_application_layer_protocol_negotiation = Integers.valueOf(ExtensionType.application_layer_protocol_negotiation);
+    public static final Integer EXT_certificate_authorities = Integers.valueOf(ExtensionType.certificate_authorities);
     public static final Integer EXT_client_certificate_type = Integers.valueOf(ExtensionType.client_certificate_type);
     public static final Integer EXT_client_certificate_url = Integers.valueOf(ExtensionType.client_certificate_url);
+    public static final Integer EXT_cookie = Integers.valueOf(ExtensionType.cookie);
+    public static final Integer EXT_early_data = Integers.valueOf(ExtensionType.early_data);
     public static final Integer EXT_ec_point_formats = Integers.valueOf(ExtensionType.ec_point_formats);
     public static final Integer EXT_encrypt_then_mac = Integers.valueOf(ExtensionType.encrypt_then_mac);
     public static final Integer EXT_extended_master_secret = Integers.valueOf(ExtensionType.extended_master_secret);
     public static final Integer EXT_heartbeat = Integers.valueOf(ExtensionType.heartbeat);
+    public static final Integer EXT_key_share = Integers.valueOf(ExtensionType.key_share);
     public static final Integer EXT_max_fragment_length = Integers.valueOf(ExtensionType.max_fragment_length);
+    public static final Integer EXT_oid_filters = Integers.valueOf(ExtensionType.oid_filters);
     public static final Integer EXT_padding = Integers.valueOf(ExtensionType.padding);
+    public static final Integer EXT_post_handshake_auth = Integers.valueOf(ExtensionType.post_handshake_auth);
+    public static final Integer EXT_pre_shared_key = Integers.valueOf(ExtensionType.pre_shared_key);
+    public static final Integer EXT_psk_key_exchange_modes = Integers.valueOf(ExtensionType.psk_key_exchange_modes);
     public static final Integer EXT_record_size_limit = Integers.valueOf(ExtensionType.record_size_limit);
     public static final Integer EXT_server_certificate_type = Integers.valueOf(ExtensionType.server_certificate_type);
     public static final Integer EXT_server_name = Integers.valueOf(ExtensionType.server_name);
@@ -49,6 +62,11 @@ public class TlsExtensionsUtils
         extensions.put(EXT_application_layer_protocol_negotiation, createALPNExtensionServer(protocolName));
     }
 
+    public static void addCertificateAuthoritiesExtension(Hashtable extensions, Vector authorities) throws IOException
+    {
+        extensions.put(EXT_certificate_authorities, createCertificateAuthoritiesExtension(authorities));
+    }
+
     public static void addClientCertificateTypeExtensionClient(Hashtable extensions, short[] certificateTypes)
         throws IOException
     {
@@ -66,6 +84,21 @@ public class TlsExtensionsUtils
         extensions.put(EXT_client_certificate_url, createClientCertificateURLExtension());
     }
 
+    public static void addCookieExtension(Hashtable extensions, byte[] cookie) throws IOException
+    {
+        extensions.put(EXT_cookie, createCookieExtension(cookie));
+    }
+
+    public static void addEarlyDataIndication(Hashtable extensions)
+    {
+        extensions.put(EXT_early_data, createEarlyDataIndication());
+    }
+
+    public static void addEarlyDataMaxSize(Hashtable extensions, long maxSize) throws IOException
+    {
+        extensions.put(EXT_early_data, createEarlyDataMaxSize(maxSize));
+    }
+
     public static void addEncryptThenMACExtension(Hashtable extensions)
     {
         extensions.put(EXT_encrypt_then_mac, createEncryptThenMACExtension());
@@ -82,18 +115,64 @@ public class TlsExtensionsUtils
         extensions.put(EXT_heartbeat, createHeartbeatExtension(heartbeatExtension));
     }
 
+    public static void addKeyShareClientHello(Hashtable extensions, Vector clientShares)
+        throws IOException
+    {
+        extensions.put(EXT_key_share, createKeyShareClientHello(clientShares));
+    }
+
+    public static void addKeyShareHelloRetryRequest(Hashtable extensions, int namedGroup)
+        throws IOException
+    {
+        extensions.put(EXT_key_share, createKeyShareHelloRetryRequest(namedGroup));
+    }
+
+    public static void addKeyShareServerHello(Hashtable extensions, KeyShareEntry serverShare)
+        throws IOException
+    {
+        extensions.put(EXT_key_share, createKeyShareServerHello(serverShare));
+    }
+
     public static void addMaxFragmentLengthExtension(Hashtable extensions, short maxFragmentLength)
         throws IOException
     {
         extensions.put(EXT_max_fragment_length, createMaxFragmentLengthExtension(maxFragmentLength));
     }
 
+    public static void addOIDFiltersExtension(Hashtable extensions, Hashtable filters) throws IOException
+    {
+        extensions.put(EXT_oid_filters, createOIDFiltersExtension(filters));
+    }
+
     public static void addPaddingExtension(Hashtable extensions, int dataLength)
         throws IOException
     {
         extensions.put(EXT_padding, createPaddingExtension(dataLength));
     }
 
+    public static void addPostHandshakeAuthExtension(Hashtable extensions)
+    {
+        extensions.put(EXT_post_handshake_auth, createPostHandshakeAuthExtension());
+    }
+
+    public static void addPreSharedKeyClientHello(Hashtable extensions, OfferedPsks offeredPsks)
+        throws IOException
+    {
+        extensions.put(EXT_pre_shared_key, createPreSharedKeyClientHello(offeredPsks));
+    }
+
+    public static void addPreSharedKeyServerHello(Hashtable extensions, int selectedIdentity)
+        throws IOException
+    {
+        extensions.put(EXT_pre_shared_key, createPreSharedKeyServerHello(selectedIdentity));
+    }
+
+    public static void addPSKKeyExchangeModesExtension(Hashtable extensions, short[] modes)
+        throws IOException
+    {
+        extensions.put(EXT_psk_key_exchange_modes, createPSKKeyExchangeModesExtension(modes));
+    }
+
     public static void addRecordSizeLimitExtension(Hashtable extensions, int recordSizeLimit)
         throws IOException
     {
@@ -112,13 +191,6 @@ public class TlsExtensionsUtils
         extensions.put(EXT_server_certificate_type, createCertificateTypeExtensionServer(certificateType));
     }
 
-    /** @deprecated Use {@link #addServerNameExtensionClient(Hashtable, Vector)} instead. */
-    public static void addServerNameExtension(Hashtable extensions, ServerNameList serverNameList)
-        throws IOException
-    {
-        extensions.put(EXT_server_name, createServerNameExtension(serverNameList));
-    }
-
     public static void addServerNameExtensionClient(Hashtable extensions, Vector serverNameList)
         throws IOException
     {
@@ -201,6 +273,12 @@ public class TlsExtensionsUtils
         return extensionData == null ? null : readALPNExtensionServer(extensionData);
     }
 
+    public static Vector getCertificateAuthoritiesExtension(Hashtable extensions) throws IOException
+    {
+        byte[] extensionData = TlsUtils.getExtensionData(extensions, EXT_certificate_authorities);
+        return extensionData == null ? null : readCertificateAuthoritiesExtension(extensionData);
+    }
+
     public static short[] getClientCertificateTypeExtensionClient(Hashtable extensions)
         throws IOException
     {
@@ -215,6 +293,19 @@ public class TlsExtensionsUtils
         return extensionData == null ? -1 : readCertificateTypeExtensionServer(extensionData);
     }
 
+    public static byte[] getCookieExtension(Hashtable extensions)
+        throws IOException
+    {
+        byte[] extensionData = TlsUtils.getExtensionData(extensions, EXT_cookie);
+        return extensionData == null ? null : readCookieExtension(extensionData);
+    }
+
+    public static long getEarlyDataMaxSize(Hashtable extensions) throws IOException
+    {
+        byte[] extensionData = TlsUtils.getExtensionData(extensions, EXT_early_data);
+        return extensionData == null ? -1L : readEarlyDataMaxSize(extensionData);
+    }
+
     public static HeartbeatExtension getHeartbeatExtension(Hashtable extensions)
         throws IOException
     {
@@ -222,6 +313,27 @@ public class TlsExtensionsUtils
         return extensionData == null ? null : readHeartbeatExtension(extensionData);
     }
 
+    public static Vector getKeyShareClientHello(Hashtable extensions)
+        throws IOException
+    {
+        byte[] extensionData = TlsUtils.getExtensionData(extensions, EXT_key_share);
+        return extensionData == null ? null : readKeyShareClientHello(extensionData);
+    }
+
+    public static int getKeyShareHelloRetryRequest(Hashtable extensions)
+        throws IOException
+    {
+        byte[] extensionData = TlsUtils.getExtensionData(extensions, EXT_key_share);
+        return extensionData == null ? -1 : readKeyShareHelloRetryRequest(extensionData);
+    }
+
+    public static KeyShareEntry getKeyShareServerHello(Hashtable extensions)
+        throws IOException
+    {
+        byte[] extensionData = TlsUtils.getExtensionData(extensions, EXT_key_share);
+        return extensionData == null ? null : readKeyShareServerHello(extensionData);
+    }
+
     public static short getMaxFragmentLengthExtension(Hashtable extensions)
         throws IOException
     {
@@ -229,6 +341,13 @@ public class TlsExtensionsUtils
         return extensionData == null ? -1 : readMaxFragmentLengthExtension(extensionData);
     }
 
+    public static Hashtable getOIDFiltersExtension(Hashtable extensions)
+        throws IOException
+    {
+        byte[] extensionData = TlsUtils.getExtensionData(extensions, EXT_oid_filters);
+        return extensionData == null ? null : readOIDFiltersExtension(extensionData);
+    }
+
     public static int getPaddingExtension(Hashtable extensions)
         throws IOException
     {
@@ -236,6 +355,27 @@ public class TlsExtensionsUtils
         return extensionData == null ? -1 : readPaddingExtension(extensionData);
     }
 
+    public static OfferedPsks getPreSharedKeyClientHello(Hashtable extensions)
+        throws IOException
+    {
+        byte[] extensionData = TlsUtils.getExtensionData(extensions, EXT_pre_shared_key);
+        return extensionData == null ? null : readPreSharedKeyClientHello(extensionData);
+    }
+
+    public static int getPreSharedKeyServerHello(Hashtable extensions)
+        throws IOException
+    {
+        byte[] extensionData = TlsUtils.getExtensionData(extensions, EXT_pre_shared_key);
+        return extensionData == null ? -1 : readPreSharedKeyServerHello(extensionData);
+    }
+
+    public static short[] getPSKKeyExchangeModesExtension(Hashtable extensions)
+        throws IOException
+    {
+        byte[] extensionData = TlsUtils.getExtensionData(extensions, EXT_psk_key_exchange_modes);
+        return extensionData == null ? null : readPSKKeyExchangeModesExtension(extensionData);
+    }
+
     public static int getRecordSizeLimitExtension(Hashtable extensions)
         throws IOException
     {
@@ -257,14 +397,6 @@ public class TlsExtensionsUtils
         return extensionData == null ? -1 : readCertificateTypeExtensionServer(extensionData);
     }
 
-    /** @deprecated Use {@link #getServerNameExtensionClient(Hashtable)} instead. */
-    public static ServerNameList getServerNameExtension(Hashtable extensions)
-        throws IOException
-    {
-        byte[] extensionData = TlsUtils.getExtensionData(extensions, EXT_server_name);
-        return extensionData == null ? null : readServerNameExtension(extensionData);
-    }
-
     public static Vector getServerNameExtensionClient(Hashtable extensions)
         throws IOException
     {
@@ -330,6 +462,12 @@ public class TlsExtensionsUtils
         return extensionData == null ? false : readClientCertificateURLExtension(extensionData);
     }
 
+    public static boolean hasEarlyDataIndication(Hashtable extensions) throws IOException
+    {
+        byte[] extensionData = TlsUtils.getExtensionData(extensions, EXT_early_data);
+        return extensionData == null ? false : readEarlyDataIndication(extensionData);
+    }
+
     public static boolean hasEncryptThenMACExtension(Hashtable extensions) throws IOException
     {
         byte[] extensionData = TlsUtils.getExtensionData(extensions, EXT_encrypt_then_mac);
@@ -349,6 +487,12 @@ public class TlsExtensionsUtils
         return extensionData == null ? false : readServerNameExtensionServer(extensionData);
     }
 
+    public static boolean hasPostHandshakeAuthExtension(Hashtable extensions) throws IOException
+    {
+        byte[] extensionData = TlsUtils.getExtensionData(extensions, EXT_post_handshake_auth);
+        return extensionData == null ? false : readPostHandshakeAuthExtension(extensionData);
+    }
+
     public static boolean hasTruncatedHMacExtension(Hashtable extensions) throws IOException
     {
         byte[] extensionData = TlsUtils.getExtensionData(extensions, EXT_truncated_hmac);
@@ -398,6 +542,32 @@ public class TlsExtensionsUtils
         return createALPNExtensionClient(protocol_name_list);
     }
 
+    public static byte[] createCertificateAuthoritiesExtension(Vector authorities) throws IOException
+    {
+        if (null == authorities || authorities.isEmpty())
+        {
+            throw new TlsFatalAlert(AlertDescription.internal_error);
+        }
+
+        ByteArrayOutputStream buf = new ByteArrayOutputStream();
+
+        // Placeholder for length
+        TlsUtils.writeUint16(0, buf);
+
+        for (int i = 0; i < authorities.size(); ++i)
+        {
+            X500Name authority = (X500Name)authorities.elementAt(i);
+            byte[] derEncoding = authority.getEncoded(ASN1Encoding.DER);
+            TlsUtils.writeOpaque16(derEncoding, buf);
+        }
+
+        int length = buf.size() - 2;
+        TlsUtils.checkUint16(length);
+        byte[] extensionData = buf.toByteArray();
+        TlsUtils.writeUint16(length, extensionData, 0);
+        return extensionData;
+    }
+
     public static byte[] createCertificateTypeExtensionClient(short[] certificateTypes) throws IOException
     {
         if (certificateTypes == null || certificateTypes.length < 1 || certificateTypes.length > 255)
@@ -418,6 +588,26 @@ public class TlsExtensionsUtils
         return createEmptyExtensionData();
     }
 
+    public static byte[] createCookieExtension(byte[] cookie) throws IOException
+    {
+        if (cookie == null || cookie.length < 1 || cookie.length >= (1 << 16))
+        {
+            throw new TlsFatalAlert(AlertDescription.internal_error);
+        }
+
+        return TlsUtils.encodeOpaque16(cookie);
+    }
+
+    public static byte[] createEarlyDataIndication()
+    {
+        return createEmptyExtensionData();
+    }
+
+    public static byte[] createEarlyDataMaxSize(long maxSize) throws IOException
+    {
+        return TlsUtils.encodeUint32(maxSize);
+    }
+
     public static byte[] createEmptyExtensionData()
     {
         return TlsUtils.EMPTY_BYTES;
@@ -448,12 +638,94 @@ public class TlsExtensionsUtils
         return buf.toByteArray();
     }
 
+    public static byte[] createKeyShareClientHello(Vector clientShares)
+        throws IOException
+    {
+        if (clientShares == null || clientShares.isEmpty())
+        {
+            return TlsUtils.encodeUint16(0);
+        }
+
+        ByteArrayOutputStream buf = new ByteArrayOutputStream();
+
+        // Placeholder for length
+        TlsUtils.writeUint16(0, buf);
+
+        for (int i = 0; i < clientShares.size(); ++i)
+        {
+            KeyShareEntry clientShare = (KeyShareEntry)clientShares.elementAt(i);
+
+            clientShare.encode(buf);
+        }
+
+        int length = buf.size() - 2;
+        TlsUtils.checkUint16(length);
+        byte[] extensionData = buf.toByteArray();
+        TlsUtils.writeUint16(length, extensionData, 0);
+        return extensionData;
+    }
+
+    public static byte[] createKeyShareHelloRetryRequest(int namedGroup)
+        throws IOException
+    {
+        return TlsUtils.encodeUint16(namedGroup);
+    }
+
+    public static byte[] createKeyShareServerHello(KeyShareEntry serverShare)
+        throws IOException
+    {
+        if (serverShare == null)
+        {
+            throw new TlsFatalAlert(AlertDescription.internal_error);
+        }
+
+        ByteArrayOutputStream buf = new ByteArrayOutputStream();
+
+        serverShare.encode(buf);
+
+        return buf.toByteArray();
+    }
+
     public static byte[] createMaxFragmentLengthExtension(short maxFragmentLength)
         throws IOException
     {
         return TlsUtils.encodeUint8(maxFragmentLength);
     }
 
+    public static byte[] createOIDFiltersExtension(Hashtable filters) throws IOException
+    {
+        ByteArrayOutputStream buf = new ByteArrayOutputStream();
+
+        // Placeholder for length
+        TlsUtils.writeUint16(0, buf);
+
+        if (null != filters)
+        {
+            Enumeration keys = filters.keys();
+            while (keys.hasMoreElements())
+            {
+                ASN1ObjectIdentifier certificateExtensionOID = (ASN1ObjectIdentifier)keys.nextElement();
+                byte[] certificateExtensionValues = (byte[])filters.get(certificateExtensionOID);
+
+                if (null == certificateExtensionOID || null == certificateExtensionValues)
+                {
+                    throw new TlsFatalAlert(AlertDescription.internal_error);
+                }
+
+                byte[] derEncoding = certificateExtensionOID.getEncoded(ASN1Encoding.DER);
+                TlsUtils.writeOpaque8(derEncoding, buf);
+
+                TlsUtils.writeOpaque16(certificateExtensionValues, buf);
+            }
+        }
+
+        int length = buf.size() - 2;
+        TlsUtils.checkUint16(length);
+        byte[] extensionData = buf.toByteArray();
+        TlsUtils.writeUint16(length, extensionData, 0);
+        return extensionData;
+    }
+
     public static byte[] createPaddingExtension(int dataLength)
         throws IOException
     {
@@ -461,6 +733,40 @@ public class TlsExtensionsUtils
         return new byte[dataLength];
     }
 
+    public static byte[] createPostHandshakeAuthExtension()
+    {
+        return createEmptyExtensionData();
+    }
+
+    public static byte[] createPreSharedKeyClientHello(OfferedPsks offeredPsks) throws IOException
+    {
+        if (offeredPsks == null)
+        {
+            throw new TlsFatalAlert(AlertDescription.internal_error);
+        }
+
+        ByteArrayOutputStream buf = new ByteArrayOutputStream();
+
+        offeredPsks.encode(buf);
+
+        return buf.toByteArray();
+    }
+
+    public static byte[] createPreSharedKeyServerHello(int selectedIdentity) throws IOException
+    {
+        return TlsUtils.encodeUint16(selectedIdentity);
+    }
+
+    public static byte[] createPSKKeyExchangeModesExtension(short[] modes) throws IOException
+    {
+        if (modes == null || modes.length < 1 || modes.length > 255)
+        {
+            throw new TlsFatalAlert(AlertDescription.internal_error);
+        }
+
+        return TlsUtils.encodeUint8ArrayWithUint8Length(modes);
+    }
+
     public static byte[] createRecordSizeLimitExtension(int recordSizeLimit)
         throws IOException
     {
@@ -472,22 +778,6 @@ public class TlsExtensionsUtils
         return TlsUtils.encodeUint16(recordSizeLimit);
     }
 
-    /** @deprecated Use {@link #createServerNameExtensionClient(Vector)} instead. */
-    public static byte[] createServerNameExtension(ServerNameList serverNameList)
-        throws IOException
-    {
-        if (serverNameList == null)
-        {
-            throw new TlsFatalAlert(AlertDescription.internal_error);
-        }
-        
-        ByteArrayOutputStream buf = new ByteArrayOutputStream();
-        
-        serverNameList.encode(buf);
-
-        return buf.toByteArray();
-    }
-
     public static byte[] createServerNameExtensionClient(Vector serverNameList)
         throws IOException
     {
@@ -679,6 +969,35 @@ public class TlsExtensionsUtils
         return (ProtocolName)protocol_name_list.elementAt(0);
     }
 
+    public static Vector readCertificateAuthoritiesExtension(byte[] extensionData) throws IOException
+    {
+        if (extensionData == null)
+        {
+            throw new IllegalArgumentException("'extensionData' cannot be null");
+        }
+        if (extensionData.length < 5)
+        {
+            throw new TlsFatalAlert(AlertDescription.decode_error);
+        }
+
+        ByteArrayInputStream buf = new ByteArrayInputStream(extensionData);
+
+        int length = TlsUtils.readUint16(buf);
+        if (length != (extensionData.length - 2))
+        {
+            throw new TlsFatalAlert(AlertDescription.decode_error);
+        }
+
+        Vector authorities = new Vector();
+        while (buf.available() > 0)
+        {
+            byte[] derEncoding = TlsUtils.readOpaque16(buf, 1);
+            ASN1Primitive asn1 = TlsUtils.readDERObject(derEncoding);
+            authorities.addElement(X500Name.getInstance(asn1));
+        }
+        return authorities;
+    }
+
     public static short[] readCertificateTypeExtensionClient(byte[] extensionData) throws IOException
     {
         short[] certificateTypes = TlsUtils.decodeUint8ArrayWithUint8Length(extensionData);
@@ -699,6 +1018,21 @@ public class TlsExtensionsUtils
         return readEmptyExtensionData(extensionData);
     }
 
+    public static byte[] readCookieExtension(byte[] extensionData) throws IOException
+    {
+        return TlsUtils.decodeOpaque16(extensionData, 1);
+    }
+
+    public static boolean readEarlyDataIndication(byte[] extensionData) throws IOException
+    {
+        return readEmptyExtensionData(extensionData);
+    }
+
+    public static long readEarlyDataMaxSize(byte[] extensionData) throws IOException
+    {
+        return TlsUtils.decodeUint32(extensionData);
+    }
+
     public static boolean readEncryptThenMACExtension(byte[] extensionData) throws IOException
     {
         return readEmptyExtensionData(extensionData);
@@ -726,12 +1060,99 @@ public class TlsExtensionsUtils
         return heartbeatExtension;
     }
 
+    public static Vector readKeyShareClientHello(byte[] extensionData)
+        throws IOException
+    {
+        if (extensionData == null)
+        {
+            throw new IllegalArgumentException("'extensionData' cannot be null");
+        }
+
+        ByteArrayInputStream buf = new ByteArrayInputStream(extensionData);
+
+        int length = TlsUtils.readUint16(buf);
+        if (length != (extensionData.length - 2))
+        {
+            throw new TlsFatalAlert(AlertDescription.decode_error);
+        }
+
+        Vector clientShares = new Vector();
+        while (buf.available() > 0)
+        {
+            KeyShareEntry clientShare = KeyShareEntry.parse(buf);
+
+            clientShares.addElement(clientShare);
+        }
+        return clientShares;
+    }
+
+    public static int readKeyShareHelloRetryRequest(byte[] extensionData)
+        throws IOException
+    {
+        return TlsUtils.decodeUint16(extensionData);
+    }
+
+    public static KeyShareEntry readKeyShareServerHello(byte[] extensionData)
+        throws IOException
+    {
+        if (extensionData == null)
+        {
+            throw new IllegalArgumentException("'extensionData' cannot be null");
+        }
+
+        ByteArrayInputStream buf = new ByteArrayInputStream(extensionData);
+
+        KeyShareEntry serverShare = KeyShareEntry.parse(buf);
+
+        TlsProtocol.assertEmpty(buf);
+
+        return serverShare;
+    }
+
     public static short readMaxFragmentLengthExtension(byte[] extensionData)
         throws IOException
     {
         return TlsUtils.decodeUint8(extensionData);
     }
 
+    public static Hashtable readOIDFiltersExtension(byte[] extensionData) throws IOException
+    {
+        if (extensionData == null)
+        {
+            throw new IllegalArgumentException("'extensionData' cannot be null");
+        }
+        if (extensionData.length < 2)
+        {
+            throw new TlsFatalAlert(AlertDescription.decode_error);
+        }
+
+        ByteArrayInputStream buf = new ByteArrayInputStream(extensionData);
+
+        int length = TlsUtils.readUint16(buf);
+        if (length != (extensionData.length - 2))
+        {
+            throw new TlsFatalAlert(AlertDescription.decode_error);
+        }
+
+        Hashtable filters = new Hashtable();
+        while (buf.available() > 0)
+        {
+            byte[] derEncoding = TlsUtils.readOpaque8(buf, 1);
+            ASN1Primitive asn1 = TlsUtils.readDERObject(derEncoding);
+            ASN1ObjectIdentifier certificateExtensionOID = ASN1ObjectIdentifier.getInstance(asn1);
+
+            if (filters.containsKey(certificateExtensionOID))
+            {
+                throw new TlsFatalAlert(AlertDescription.illegal_parameter);
+            }
+
+            byte[] certificateExtensionValues = TlsUtils.readOpaque16(buf);
+
+            filters.put(certificateExtensionOID, certificateExtensionValues);
+        }
+        return filters;
+    }
+
     public static int readPaddingExtension(byte[] extensionData)
         throws IOException
     {
@@ -749,6 +1170,42 @@ public class TlsExtensionsUtils
         return extensionData.length;
     }
 
+    public static boolean readPostHandshakeAuthExtension(byte[] extensionData) throws IOException
+    {
+        return readEmptyExtensionData(extensionData);
+    }
+
+    public static OfferedPsks readPreSharedKeyClientHello(byte[] extensionData) throws IOException
+    {
+        if (extensionData == null)
+        {
+            throw new IllegalArgumentException("'extensionData' cannot be null");
+        }
+
+        ByteArrayInputStream buf = new ByteArrayInputStream(extensionData);
+
+        OfferedPsks offeredPsks = OfferedPsks.parse(buf);
+
+        TlsProtocol.assertEmpty(buf);
+
+        return offeredPsks;
+    }
+
+    public static int readPreSharedKeyServerHello(byte[] extensionData) throws IOException
+    {
+        return TlsUtils.decodeUint16(extensionData);
+    }
+
+    public static short[] readPSKKeyExchangeModesExtension(byte[] extensionData) throws IOException
+    {
+        short[] modes = TlsUtils.decodeUint8ArrayWithUint8Length(extensionData);
+        if (modes.length < 1)
+        {
+            throw new TlsFatalAlert(AlertDescription.decode_error);
+        }
+        return modes;
+    }
+
     public static int readRecordSizeLimitExtension(byte[] extensionData)
         throws IOException
     {
@@ -761,24 +1218,6 @@ public class TlsExtensionsUtils
         return recordSizeLimit;
     }
 
-    /** @deprecated Use {@link #readServerNameExtensionClient(byte[])} instead. */
-    public static ServerNameList readServerNameExtension(byte[] extensionData)
-        throws IOException
-    {
-        if (extensionData == null)
-        {
-            throw new IllegalArgumentException("'extensionData' cannot be null");
-        }
-
-        ByteArrayInputStream buf = new ByteArrayInputStream(extensionData);
-
-        ServerNameList serverNameList = ServerNameList.parse(buf);
-
-        TlsProtocol.assertEmpty(buf);
-
-        return serverNameList;
-    }
-
     public static Vector readServerNameExtensionClient(byte[] extensionData)
         throws IOException
     {
