@@ -24,9 +24,9 @@ import org.bouncycastle.tls.SignatureAndHashAlgorithm;
 import org.bouncycastle.tls.TlsAuthentication;
 import org.bouncycastle.tls.TlsCredentialedSigner;
 import org.bouncycastle.tls.TlsCredentials;
+import org.bouncycastle.tls.TlsExtensionsUtils;
 import org.bouncycastle.tls.TlsFatalAlert;
 import org.bouncycastle.tls.TlsServerCertificate;
-import org.bouncycastle.tls.TlsUtils;
 import org.bouncycastle.tls.crypto.TlsCertificate;
 import org.bouncycastle.tls.crypto.TlsCrypto;
 import org.bouncycastle.tls.crypto.TlsStreamSigner;
@@ -78,23 +78,21 @@ class TlsTestClientImpl
         }
     }
 
-    public ProtocolVersion[] getSupportedVersions()
-    {
-        if (null != config.clientSupportedVersions)
-        {
-            return config.clientSupportedVersions;
-        }
-
-        return super.getSupportedVersions();
-    }
-
     public Hashtable getClientExtensions() throws IOException
     {
         Hashtable clientExtensions = super.getClientExtensions();
-        if (clientExtensions != null && !config.clientSendSignatureAlgorithms)
+        if (clientExtensions != null)
         {
-            clientExtensions.remove(TlsUtils.EXT_signature_algorithms);
-            this.supportedSignatureAlgorithms = null;
+            if (!config.clientSendSignatureAlgorithms)
+            {
+                clientExtensions.remove(TlsExtensionsUtils.EXT_signature_algorithms);
+                this.supportedSignatureAlgorithms = null;
+            }
+            if (!config.clientSendSignatureAlgorithmsCert)
+            {
+                clientExtensions.remove(TlsExtensionsUtils.EXT_signature_algorithms_cert);
+                this.supportedSignatureAlgorithmsCert = null;
+            }
         }
         return clientExtensions;
     }
@@ -317,6 +315,16 @@ class TlsTestClientImpl
         return bs;
     }
 
+    protected ProtocolVersion[] getSupportedVersions()
+    {
+        if (null != config.clientSupportedVersions)
+        {
+            return config.clientSupportedVersions;
+        }
+
+        return super.getSupportedVersions();
+    }
+
     protected String hex(byte[] data)
     {
         return data == null ? "(null)" : Hex.toHexString(data);
